import { __assign } from "tslib";
function topologicalOrderCommits(commits, commitsMap) {
    // sort commits by committer date from latest to oldest
    var commitsSortByCommitterDate = commits.sort(function (a, b) { return b.committerDate.getTime() - a.committerDate.getTime(); });
    var sortedCommits = [];
    var seen = new Map();
    function dfs(commit) {
        var commitHash = commit.hash;
        if (seen.get(commitHash)) {
            return;
        }
        seen.set(commitHash, true);
        commit.children.forEach(function (children) {
            dfs(commitsMap.get(children));
        });
        sortedCommits.push(commitHash);
    }
    // compute topological order of commits
    commitsSortByCommitterDate.forEach(function (commit) {
        dfs(commit);
    });
    return sortedCommits;
}
function computeColumns(orderedCommitHashes, commitsMap) {
    // Branch positions organized as columns, each column can have multiple branches
    var columns = [];
    // Column position of each commit
    var commitYs = new Map();
    var commitXs = new Map(orderedCommitHashes.map(function (commitHash, index) { return [commitHash, index]; }));
    function updateColumns(col, end, endCommitHash) {
        columns[col][columns[col].length - 1] = __assign(__assign({}, columns[col][columns[col].length - 1]), { end: end, endCommitHash: endCommitHash });
    }
    // branch order is used to determine the color of the branch
    var branchOrder = 0;
    // Compute column position of each commit by topological order
    // So the child commit will always be computed before its parent commit
    orderedCommitHashes.forEach(function (commitHash, index) {
        var commit = commitsMap.get(commitHash);
        var branchChildren = commit.children.filter(function (child) { return commitsMap.get(child).parents[0] === commit.hash; });
        var isLastCommitOnBranch = commit.children.length === 0;
        var isChildOfNonMergeCommit = branchChildren.length > 0;
        var commitY = -1;
        // for paint branch path
        var isFirstCommit = commit.parents.length === 0;
        // Update columns based on the commit type
        if (isLastCommitOnBranch) {
            // Put commit as a new column
            columns.push([
                {
                    start: index,
                    end: isFirstCommit ? index : Infinity,
                    endCommitHash: commit.hash,
                    branchOrder: branchOrder,
                },
            ]);
            branchOrder++;
            commitY = columns.length - 1;
        }
        else if (isChildOfNonMergeCommit) {
            // Put commit to its left most column of its children (non-merge commit)
            var branchChildrenYs = branchChildren
                .map(function (childHash) { return commitYs.get(childHash); })
                .filter(function (y) { return y !== undefined; });
            // Set the commit to the left most column of its children
            commitY = Math.min.apply(Math, branchChildrenYs);
            // Update the left most column
            updateColumns(commitY, isFirstCommit ? index : Infinity, commit.hash);
            // Update other columns with the commit as the end commit
            branchChildrenYs
                .filter(function (childY) { return childY !== commitY; })
                .forEach(function (childY) {
                updateColumns(childY, index - 1, commit.hash);
            });
        }
        else {
            // Find a column so the commit can connect to its children (merge commit) without overlapping with existing branches on columns
            // Otherwise, put the commit as a new column
            // Highest pos of child commit
            var minChildX_1 = Infinity;
            // Right most pos of child commit
            var maxChildY_1 = -1;
            commit.children.forEach(function (child) {
                var childX = commitXs.get(child);
                var childY = commitYs.get(child);
                if (childX < minChildX_1) {
                    minChildX_1 = childX;
                }
                if (childY > maxChildY_1) {
                    maxChildY_1 = childY;
                }
            });
            var colFitAtEnd = columns.slice(maxChildY_1 + 1).findIndex(function (column) {
                return minChildX_1 >= column[column.length - 1].end;
            });
            var col = colFitAtEnd === -1 ? -1 : maxChildY_1 + 1 + colFitAtEnd;
            if (col === -1) {
                // minChildX + 1 so it will be extended by the merge curve
                columns.push([
                    {
                        start: minChildX_1 + 1,
                        end: isFirstCommit ? index : Infinity,
                        endCommitHash: commit.hash,
                        branchOrder: branchOrder,
                    },
                ]);
                branchOrder++;
                commitY = columns.length - 1;
            }
            else {
                commitY = col;
                columns[col].push({
                    start: minChildX_1 + 1,
                    end: isFirstCommit ? index : Infinity,
                    endCommitHash: commit.hash,
                    branchOrder: branchOrder,
                });
                branchOrder++;
            }
        }
        commitYs.set(commitHash, commitY);
    });
    return { columns: columns, commitYs: commitYs };
}
export function computePosition(commits) {
    var commitsMap = new Map(commits.map(function (commit) { return [commit.hash, commit]; }));
    var orderedCommitHashes = topologicalOrderCommits(commits, commitsMap);
    var _a = computeColumns(orderedCommitHashes, commitsMap), columns = _a.columns, commitYs = _a.commitYs;
    // To be compatible with exisiting data structure used in render
    // Add position info to commits and end commit to columns
    var commitsMapWithPos = new Map(orderedCommitHashes.map(function (commitHash, ix) { return [
        commitHash,
        __assign(__assign({}, commitsMap.get(commitHash)), { x: ix, y: commitYs.get(commitHash) }),
    ]; }));
    var columnsWithEndCommit = columns.map(function (column) {
        return column.map(function (branchPath) {
            return __assign(__assign({}, branchPath), { endCommit: commitsMapWithPos.get(branchPath.endCommitHash) });
        });
    });
    return { columns: columnsWithEndCommit, commitsMap: commitsMapWithPos };
}
//# sourceMappingURL=computePosition.js.map